/* Copyright (C) 2004-2005 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lislib.h"

int lis_vector_create(int local_n, int global_n, LIS_Comm comm, LIS_VECTOR *vec)
{
	int n,gn;
	int nprocs,my_rank;
	int is,ie;
	int i;
	int *ranges;

	*vec = NULL;

	if( global_n>0 && local_n>global_n )
	{
		SETERR("local n(=%d) is larger than global n(=%d)\n",local_n,global_n);
		return LIS_FAILS;
	}
	if( local_n<0 || global_n<0 )
	{
		SETERR("local n(=%d) or global n(=%d) are less than 0\n",local_n,global_n);
		return LIS_FAILS;
	}
	if( local_n==0 && global_n==0 )
	{
		SETERR("local n(=%d) and global n(=%d) are 0\n",local_n,global_n);
		return LIS_FAILS;
	}

	*vec = (LIS_VECTOR)lis_malloc( sizeof(struct LIS_VECTOR_STRUCT) );
	if( NULL==*vec )
	{
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	#ifdef USE_MPI
		MPI_Comm_size(comm,&nprocs);
		MPI_Comm_rank(comm,&my_rank);
		ranges = (int *)lis_malloc( (nprocs+1)*sizeof(int) );
		if( ranges==NULL )
		{
			lis_vector_destroy(*vec);
			*vec = NULL;
			SETERR("out of memory\n");
			return LIS_OUT_OF_MEMORY;
		}
	#else
		nprocs  = 1;
		my_rank = 0;
		ranges  = NULL;
	#endif

	if( local_n==0 )
	{
		#ifdef USE_MPI
			LIS_GET_ISIE(my_rank,nprocs,global_n,is,ie);
			local_n = ie-is;
			MPI_Allgather(&ie,1,MPI_INT,&ranges[1],1,MPI_INT,comm);
			ranges[0] = 0;
		#else
			local_n = global_n;
			is      = 0;
			ie      = global_n;
		#endif
	}
		else
		{
		#ifdef USE_MPI
			MPI_Allgather(&local_n,1,MPI_INT,&ranges[1],1,MPI_INT,comm);
			ranges[0] = 0;
			for(i=0;i<nprocs;i++)
			{
				ranges[i+1] += ranges[i];
			}
			global_n = ranges[nprocs];
		#else
			global_n = local_n;
		#endif
	}

	(*vec)->ranges      = ranges;
	(*vec)->value = (LIS_SCALAR *)lis_malloc( local_n*sizeof(LIS_SCALAR) );
	if( NULL==(*vec)->value )
	{
		lis_vector_destroy(*vec);
		*vec = NULL;
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	(*vec)->g_index     = NULL;

	(*vec)->n           = local_n;
	(*vec)->gn          = global_n;
	(*vec)->np          = local_n;
	(*vec)->comm        = comm;
	(*vec)->my_rank     = my_rank;
	(*vec)->nprocs      = nprocs;
	(*vec)->is          = is;
	(*vec)->ie          = ie;

	return LIS_SUCCESS;
}

int lis_work_vector_create(int local_n, int global_n, LIS_Comm comm, LIS_VECTOR *vec)
{
	int n,gn;
	int nprocs,my_rank;
	int is,ie;
	int i;
	int *ranges;

	*vec = NULL;

	if( global_n>0 && local_n>global_n )
	{
		SETERR("local n(=%d) is larger than global n(=%d)\n",local_n,global_n);
		return LIS_FAILS;
	}
	if( local_n<0 || global_n<0 )
	{
		SETERR("local n(=%d) or global n(=%d) are less than 0\n",local_n,global_n);
		return LIS_FAILS;
	}
	if( local_n==0 && global_n==0 )
	{
		SETERR("local n(=%d) and global n(=%d) are 0\n",local_n,global_n);
		return LIS_FAILS;
	}

	*vec = (LIS_VECTOR)lis_malloc( sizeof(struct LIS_VECTOR_STRUCT) );
	if( NULL==*vec )
	{
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	#ifdef USE_MPI
		MPI_Comm_size(comm,&nprocs);
		MPI_Comm_rank(comm,&my_rank);
		ranges = (int *)lis_malloc( (nprocs+1)*sizeof(int) );
		if( ranges==NULL )
		{
			lis_vector_destroy(*vec);
			*vec = NULL;
			SETERR("out of memory\n");
			return LIS_OUT_OF_MEMORY;
		}
	#else
		nprocs  = 1;
		my_rank = 0;
		ranges  = NULL;
	#endif

	if( local_n==0 )
	{
		#ifdef USE_MPI
			LIS_GET_ISIE(my_rank,nprocs,global_n,is,ie);
			local_n = ie-is;
			MPI_Allgather(&ie,1,MPI_INT,&ranges[1],1,MPI_INT,comm);
			ranges[0] = 0;
		#else
			local_n = global_n;
			is      = 0;
			ie      = global_n;
		#endif
	}
		else
		{
		#ifdef USE_MPI
			MPI_Allgather(&local_n,1,MPI_INT,&ranges[1],1,MPI_INT,comm);
			ranges[0] = 0;
			for(i=0;i<nprocs;i++)
			{
				ranges[i+1] += ranges[i];
			}
			global_n = ranges[nprocs];
		#else
			global_n = local_n;
		#endif
	}

	(*vec)->ranges      = ranges;
	(*vec)->value = (LIS_SCALAR *)lis_malloc( local_n*sizeof(LIS_SCALAR) );
	if( NULL==(*vec)->value )
	{
		lis_vector_destroy(*vec);
		*vec = NULL;
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}

	(*vec)->g_index     = NULL;

	(*vec)->n           = local_n;
	(*vec)->gn          = global_n;
	(*vec)->np          = local_n;
	(*vec)->comm        = comm;
	(*vec)->my_rank     = my_rank;
	(*vec)->nprocs      = nprocs;
	(*vec)->is          = is;
	(*vec)->ie          = ie;

	return LIS_SUCCESS;
}

int lis_vector_get_range(LIS_VECTOR v, int *is, int *ie)
{
	if( v==NULL )
	{
		SETERR("vector v is undefined\n");
		return LIS_FAILS;
	}
	*is = v->is;
	*ie = v->ie;

	return LIS_SUCCESS;
}

int lis_vector_set(int i, LIS_SCALAR value, LIS_VECTOR v)
{
	int n,gn,is,ie;

	n   = v->n;
	gn  = v->gn;
	is  = v->is;
	ie  = v->ie;
	if( i<is || i>=ie )
	{
		SETERR("lis_vector_set: i(=%d) is less than %d or larger than %d\n",i,is,ie);
		return LIS_FAILS;
	}

	v->value[i-is] = value;
	return LIS_SUCCESS;
}

int lis_vector_duplicate(LIS_VECTOR vin, LIS_VECTOR *vout)
{
	int n,np;
	int nprocs;
	int is,ie;
	int i;
	int *ranges;
	LIS_SCALAR *value;

	nprocs = vin->nprocs;
	np     = vin->np;
	*vout = NULL;
	*vout = (LIS_VECTOR)lis_malloc( sizeof(struct LIS_VECTOR_STRUCT) );
	if( NULL==*vout )
	{
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}
	value = (LIS_SCALAR *)lis_malloc( np*sizeof(LIS_SCALAR) );
	if( NULL==value )
	{
		lis_vector_destroy(*vout);
		*vout = NULL;
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}
	(*vout)->value = value;

	#ifdef USE_MPI
		ranges = (int *)lis_malloc( (nprocs+1)*sizeof(int) );
		if( ranges==NULL )
		{
			lis_vector_destroy(*vout);
			*vout = NULL;
			SETERR("out of memory\n");
			return LIS_OUT_OF_MEMORY;
		}
		for(i=0;i<nprocs+1;i++) ranges[i] = vin->ranges[i];
		(*vout)->ranges      = ranges;
	#else
		(*vout)->ranges      = NULL;
	#endif

	(*vout)->g_index     = vin->g_index;
	(*vout)->n           = vin->n;
	(*vout)->gn          = vin->gn;
	(*vout)->np          = vin->np;
	(*vout)->comm        = vin->comm;
	(*vout)->my_rank     = vin->my_rank;
	(*vout)->nprocs      = vin->nprocs;
	(*vout)->is          = vin->is;
	(*vout)->ie          = vin->ie;

	return LIS_SUCCESS;
}

int lis_vector_duplicateM(LIS_MATRIX A, LIS_VECTOR *vout)
{
	int n,np;
	int nprocs;
	int is,ie;
	int i;
	int *ranges;
	LIS_SCALAR *value;

	nprocs = A->nprocs;
	np     = A->np;
	*vout = NULL;
	*vout = (LIS_VECTOR)lis_malloc( sizeof(struct LIS_VECTOR_STRUCT) );
	if( NULL==*vout )
	{
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}
	value = (LIS_SCALAR *)lis_malloc( np*sizeof(LIS_SCALAR) );
	if( NULL==value )
	{
		lis_vector_destroy(*vout);
		*vout = NULL;
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}
	(*vout)->value = value;

	#ifdef USE_MPI
		ranges = (int *)lis_malloc( (nprocs+1)*sizeof(int) );
		if( ranges==NULL )
		{
			lis_vector_destroy(*vout);
			*vout = NULL;
			SETERR("out of memory\n");
			return LIS_OUT_OF_MEMORY;
		}
		for(i=0;i<nprocs+1;i++) ranges[i] = A->ranges[i];
		(*vout)->ranges      = ranges;
	#else
		(*vout)->ranges      = NULL;
	#endif

	(*vout)->g_index     = NULL;
//	(*vout)->g_index     = A->g_index;
	(*vout)->n           = A->n;
	(*vout)->gn          = A->gn;
	(*vout)->np          = A->np;
	(*vout)->comm        = A->comm;
	(*vout)->my_rank     = A->my_rank;
	(*vout)->nprocs      = A->nprocs;
	(*vout)->is          = A->is;
	(*vout)->ie          = A->ie;

	return LIS_SUCCESS;
}

int lis_vector_destroy(LIS_VECTOR vec)
{
	if( vec )
	{
		if( vec->value ) lis_free( vec->value );
		if( vec->ranges ) lis_free( vec->ranges );
		if( vec->g_index ) lis_free( vec->g_index );
		lis_free(vec);
	}
	return LIS_SUCCESS;
}

int lis_vector_scale(LIS_SCALAR alpha, LIS_VECTOR vx)
{
	int i,n;
	LIS_SCALAR *x;

	x = vx->value;
	n = vx->n;
#pragma omp parallel for private(i)
	for(i=0; i<n; i++)
	{
		x[i] = alpha * x[i];
	}
	return LIS_SUCCESS;
}

int lis_vector_set_all(LIS_SCALAR alpha, LIS_VECTOR vx)
{
	int i,n;
	LIS_SCALAR *x;

	x = vx->value;
	n = vx->n;
#pragma omp parallel for private(i)
	for(i=0; i<n; i++)
	{
		x[i] = alpha;
	}
	return LIS_SUCCESS;
}

int lis_vector_copy(LIS_VECTOR vx, LIS_VECTOR vy)
{
	int i,n;
	LIS_SCALAR *x,*y;

	x = vx->value;
	y = vy->value;
	n = vx->n;
	if( n!=vy->n )
	{
		SETERR("length of vector x and y is not equal\n");
		return LIS_FAILS;
	}
#pragma omp parallel for private(i)
	for(i=0; i<n; i++)
	{
		y[i] = x[i];
	}
	return LIS_SUCCESS;
}

int lis_vector_xpay(LIS_VECTOR vx, LIS_SCALAR alpha, LIS_VECTOR vy)
{
	int i,n;
	LIS_SCALAR *x,*y;

	x    = vx->value;
	y    = vy->value;
	n    = vx->n;
	if( n!=vy->n )
	{
		SETERR("length of vector x and y is not equal\n");
		return LIS_FAILS;
	}

#pragma omp parallel for private(i)
	for(i=0; i<n; i++)
	{
		y[i] = x[i] + alpha * y[i];
	}
	return LIS_SUCCESS;
}

int lis_vector_axpy(LIS_SCALAR alpha, LIS_VECTOR vx, LIS_VECTOR vy)
{
	int i,n;
	LIS_SCALAR *x,*y;

	x    = vx->value;
	y    = vy->value;
	n    = vx->n;
	if( n!=vy->n )
	{
		SETERR("length of vector x and y is not equal\n");
		return LIS_FAILS;
	}

#pragma omp parallel for private(i)
	for(i=0; i<n; i++)
	{
		y[i] += alpha * x[i];
	}
	return LIS_SUCCESS;
}

int lis_vector_axpyz(LIS_SCALAR alpha, LIS_VECTOR vx, LIS_VECTOR vy, LIS_VECTOR vz)
{
	int i,n;
	LIS_SCALAR *x,*y,*z;

	x    = vx->value;
	y    = vy->value;
	z    = vz->value;
	n    = vx->n;
	if( n!=vy->n || n!=vz->n )
	{
		SETERR("length of vector x, y, and z is not equal\n");
		return LIS_FAILS;
	}

#pragma omp parallel for private(i)
	for(i=0; i<n; i++)
	{
		z[i] = alpha * x[i] + y[i];
	}
	return LIS_SUCCESS;
}

int lis_vector_dot(LIS_VECTOR vx, LIS_VECTOR vy, LIS_SCALAR *val)
{
#ifdef _OPENMP
	int i,k,n,nprocs,my_rank;
	LIS_SCALAR t;
	LIS_SCALAR dot = (LIS_SCALAR)0;
	LIS_SCALAR *tmp;
	LIS_SCALAR *x,*y;

	x = vx->value;
	y = vy->value;
	n = vx->n;
	if( n!=vy->n )
	{
		SETERR("length of vector x and y is not equal\n");
		return LIS_FAILS;
	}
#define PADD 128
	nprocs = omp_get_max_threads();
	tmp = (LIS_SCALAR *)lis_malloc( nprocs*PADD*sizeof(LIS_SCALAR) );
#pragma omp parallel private(i,t,my_rank)
{
	my_rank = omp_get_thread_num();
	t       = (LIS_SCALAR)0.0;
#pragma omp for
	for(i=0; i<n; i++)
	{
		t += x[i]*y[i];
	}
	tmp[my_rank*PADD] = t;
}
	for(i=0;i<nprocs;i++)
	{
		dot += tmp[i*PADD];
	}
	lis_free(tmp);
	*val = dot;
	return LIS_SUCCESS;
#else
#ifdef USE_MPI
	int i,n;
	LIS_SCALAR dot = (LIS_SCALAR)0;
	LIS_SCALAR ret;
	LIS_SCALAR *x,*y;
	MPI_Comm   comm;

	x    = vx->value;
	y    = vy->value;
	n    = vx->n;
	comm = vx->comm;
	if( n!=vy->n )
	{
		SETERR("length of vector x and y is not equal\n");
		return LIS_FAILS;
	}

	for(i=0; i<n; i++)
	{
		dot += x[i]*y[i];
	}
	MPI_Allreduce(&dot,&ret,1,MPI_DOUBLE,MPI_SUM,comm);
	*val = ret;
	return LIS_SUCCESS;
#else
	int i,n;
	LIS_SCALAR dot = (LIS_SCALAR)0;
	LIS_SCALAR *x,*y;

	x    = vx->value;
	y    = vy->value;
	n    = vx->n;
	if( n!=vy->n )
	{
		SETERR("length of vector x and y is not equal\n");
		return LIS_FAILS;
	}

	for(i=0; i<n; i++)
	{
		dot += x[i]*y[i];
	}
	*val = dot;
	return LIS_SUCCESS;
#endif
#endif
}

int lis_vector_nrm2(LIS_VECTOR vx, LIS_SCALAR *val)
{
#ifdef _OPENMP
	int i,k,n,nprocs,my_rank;
	LIS_SCALAR t;
	LIS_SCALAR dot = (LIS_SCALAR)0;
	LIS_SCALAR *tmp;
	LIS_SCALAR *x;

	x = vx->value;
	n = vx->n;

#define PADD 128
	nprocs = omp_get_max_threads();
	tmp = (LIS_SCALAR *)lis_malloc( nprocs*PADD*sizeof(LIS_SCALAR) );
#pragma omp parallel private(i,t,my_rank)
{
	my_rank = omp_get_thread_num();
	t       = (LIS_SCALAR)0.0;
#pragma omp for
	for(i=0; i<n; i++)
	{
		t += x[i]*x[i];
	}
	tmp[my_rank*PADD] = t;
}
	for(i=0;i<nprocs;i++)
	{
		dot += tmp[i*PADD];
	}
	lis_free(tmp);
	*val = sqrt(dot);
	return LIS_SUCCESS;
#else
#ifdef USE_MPI
	int i,n;
	LIS_SCALAR dot = (LIS_SCALAR)0;
	LIS_SCALAR ret;
	LIS_SCALAR *x;
	MPI_Comm   comm;

	x    = vx->value;
	n    = vx->n;
	comm = vx->comm;

	for(i=0; i<n; i++)
	{
		dot += x[i]*x[i];
	}
	MPI_Allreduce(&dot,&ret,1,MPI_DOUBLE,MPI_SUM,comm);
	*val = sqrt(ret);
	return LIS_SUCCESS;
#else
	int i,n;
	LIS_SCALAR dot;
	LIS_SCALAR *x;

	x    = vx->value;
	n    = vx->n;
	dot  = 0;

	for(i=0;i<n;i++)
	{
		dot += x[i]*x[i];
	}
	*val = sqrt(dot);
	return LIS_SUCCESS;
#endif
#endif
}
 

int lis_vector_nrm2_s2(LIS_VECTOR vx, LIS_SCALAR *val)
{
	int i,n;
	LIS_SCALAR dot,t;
	LIS_SCALAR t1,t2;
	LIS_SCALAR s1,s2;
	LIS_SCALAR p1,p2;
	LIS_SCALAR *x;

	x    = vx->value;
	n    = vx->n;

	if( n<2 )
	{
		dot = 0;
		i   = 0;
	}
	else
	{
		s1 = s2 = 0;
		t1 = *(x++);
		p1 = t1*t1;
		for(i=2; i<n; i+=2)
		{
			t2  = *(x++);
			p2  = t2*t2;
			s1 += p1;
			t1  = *(x++);
			p1  = t1*t1;
			s2 += p2;
		}
		s1 += p1;
		dot = s1+s2;
		i-=1;
	}
	for(;i<n;i++)
	{
		t    = *(x++);
		dot += t*t;
	}
	*val = sqrt(dot);
	return LIS_SUCCESS;
}

int lis_vector_nrm2_s3(LIS_VECTOR vx, LIS_SCALAR *val)
{
	int i,n;
	LIS_SCALAR dot,t;
	LIS_SCALAR t1,t2,t3;
	LIS_SCALAR s1,s2,s3;
	LIS_SCALAR p1,p2,p3;
	LIS_SCALAR *x;

	x    = vx->value;
	n    = vx->n;

	if( n<3 )
	{
		dot = 0;
		i   = 0;
	}
	else
	{
		s1 = s2 = s3 = 0;
		t1 = *(x++);
		t2 = *(x++);
		p1 = t1*t1;
		for(i=4; i<n; i+=3)
		{
			t3  = *(x++);
			p2  = t2*t2;
			s1 += p1;

			t1  = *(x++);
			p3  = t3*t3;
			s2 += p2;

			t2  = *(x++);
			p1  = t1*t1;
			s3 += p3;
		}
		p2  = t2*t2;
		s1 += p1;
		s2 += p2;
		dot = s1+s2+s3;
		i-=2;
	}
	for(;i<n;i++)
	{
		t    = *(x++);
		dot += t*t;
	}
	*val = sqrt(dot);
	return LIS_SUCCESS;
}

int lis_vector_nrm2_s4(LIS_VECTOR vx, LIS_SCALAR *val)
{
	int i,n;
	LIS_SCALAR dot,t;
	LIS_SCALAR t1,t2,t3,t4;
	LIS_SCALAR s1,s2,s3,s4;
	LIS_SCALAR p1,p2,p3,p4;
	LIS_SCALAR *x;

	x    = vx->value;
	n    = vx->n;

	if( n<4 )
	{
		dot = 0;
		i   = 0;
	}
	else
	{
		s1 = s2 = s3 = s4 = 0;
		t1 = *(x++);
		t2 = *(x++);
		p1 = t1*t1;
		for(i=5; i<n; i+=4)
		{
			t3  = *(x++);
			p2  = t2*t2;
			s1 += p1;

			t4  = *(x++);
			p3  = t3*t3;
			s2 += p2;

			t1  = *(x++);
			p4  = t4*t4;
			s3 += p3;

			t2  = *(x++);
			p1  = t1*t1;
			s4 += p4;
		}
		p2  = t2*t2;
		s1 += p1;
		s2 += p2;
		dot = s1+s2+s3+s4;
		i-=3;
	}
	for(;i<n;i++)
	{
		t    = *(x++);
		dot += t*t;
	}
	*val = sqrt(dot);
	return LIS_SUCCESS;
}

int lis_vector_nrm2_s5(LIS_VECTOR vx, LIS_SCALAR *val)
{
	int i,n;
	LIS_SCALAR dot,t;
	LIS_SCALAR t1,t2,t3,t4,t5;
	LIS_SCALAR s1,s2,s3,s4,s5;
	LIS_SCALAR p1,p2,p3,p4,p5;
	LIS_SCALAR *x;

	x    = vx->value;
	n    = vx->n;

	if( n<5 )
	{
		dot = 0;
		i   = 0;
	}
	else
	{
		s1 = s2 = s3 = s4 = s5 = 0;
		t1 = *(x++);
		t2 = *(x++);
		t3 = *(x++);
		p1 = t1*t1;
		for(i=7; i<n; i+=5)
		{
			t4  = *(x++);
			p2  = t2*t2;
			s1 += p1;

			t5  = *(x++);
			p3  = t3*t3;
			s2 += p2;

			t1  = *(x++);
			p4  = t4*t4;
			s3 += p3;

			t2  = *(x++);
			p5  = t5*t5;
			s4 += p4;

			t3  = *(x++);
			p1  = t1*t1;
			s5 += p5;
		}
		s1 += p1;
		p2  = t2*t2;
		s2 += p2;
		p3  = t3*t3;
		s3 += p3;

		dot = s1+s2+s3+s4+s5;
		i-=4;
	}
	for(;i<n;i++)
	{
		t    = *(x++);
		dot += t*t;
	}
	*val = sqrt(dot);
	return LIS_SUCCESS;
}

int lis_vector_nrm2_s6(LIS_VECTOR vx, LIS_SCALAR *val)
{
	int i,n;
	LIS_SCALAR dot,t;
	LIS_SCALAR t1,t2,t3,t4,t5,t6;
	LIS_SCALAR s1,s2,s3,s4,s5,s6;
	LIS_SCALAR p1,p2,p3,p4,p5,p6;
	LIS_SCALAR *x;

	x    = vx->value;
	n    = vx->n;

	if( n<5 )
	{
		dot = 0;
		i   = 0;
	}
	else
	{
		s1 = s2 = s3 = s4 = s5 = s6 = 0;
		t1 = *(x++);
		t2 = *(x++);
		t3 = *(x++);
		p1 = t1*t1;
		p2 = t2*t2;
		for(i=8; i<n; i+=6)
		{
			t4  = *(x++);
			p3  = t3*t3;
			s1 += p1;

			t5  = *(x++);
			p4  = t4*t4;
			s2 += p2;

			t6  = *(x++);
			p5  = t5*t5;
			s3 += p3;

			t1  = *(x++);
			p6  = t6*t6;
			s4 += p4;

			t2  = *(x++);
			p1  = t1*t1;
			s5 += p5;

			t3  = *(x++);
			p2  = t2*t2;
			s6 += p6;
		}
		s1 += p1;
		p2  = t2*t2;
		s2 += p2;
		p3  = t3*t3;
		s3 += p3;

		dot = s1+s2+s3+s4+s5+s6;
		i-=5;
	}
	for(;i<n;i++)
	{
		t    = *(x++);
		dot += t*t;
	}
	*val = sqrt(dot);
	return LIS_SUCCESS;
}

int lis_vector_nrm2_u4f(LIS_VECTOR vx, LIS_SCALAR *val)
{
	int i,n;
	LIS_SCALAR dot,t;
	LIS_SCALAR t1,t2,t3,t4;
	LIS_SCALAR s1,s2,s3,s4;
	LIS_SCALAR *x;

	x    = vx->value;
	n    = vx->n;

	if( n<4 )
	{
		dot = 0;
		i   = 0;
	}
	else
	{
		s1 = s2 = s3 = s4 = 0;
		t1 = *(x++);
		t2 = *(x++);
		for(i=5; i<n; i+=4)
		{
			t3  = *(x++);
			s1 += t1*t1;

			t4  = *(x++);
			s2 += t2*t2;

			t1  = *(x++);
			s3 += t3*t3;

			t2  = *(x++);
			s4 += t4*t4;
		}
		s1 += t1*t1;
		s2 += t2*t2;
		dot = s1+s2+s3+s4;
		i-=3;
	}
	for(;i<n;i++)
	{
		t    = *(x++);
		dot += t*t;
	}
	*val = sqrt(dot);
	return LIS_SUCCESS;
}


int lis_vector_nrm2_u5(LIS_VECTOR vx, LIS_SCALAR *val)
{
	int i,n;
	LIS_SCALAR dot,t;
	LIS_SCALAR t1,t2,t3,t4,t5;
	LIS_SCALAR s1,s2,s3,s4,s5;
	LIS_SCALAR *x;

	x    = vx->value;
	n    = vx->n;

	for(i=4; i<n; i+=5)
	{
		t1  = *(x++);
		s1 += t1*t1;

		t2  = *(x++);
		s2 += t2*t2;

		t3  = *(x++);
		s3 += t3*t3;

		t4  = *(x++);
		s4 += t4*t4;

		t5  = *(x++);
		s5 += t5*t5;
	}
	dot = s1+s2+s3+s4+s5;
	for(i-=4;i<n;i++)
	{
		t    = *(x++);
		dot += t*t;
	}
	*val = sqrt(dot);
	return LIS_SUCCESS;
}

int lis_vector_nrm2_u5p(LIS_VECTOR vx, LIS_SCALAR *val)
{
	int i,n;
	LIS_SCALAR dot,t;
	LIS_SCALAR t1,t2,t3,t4,t5;
	LIS_SCALAR s1,s2,s3,s4,s5;
	LIS_SCALAR p1,p2,p3,p4,p5;
	LIS_SCALAR *x;

	x    = vx->value;
	n    = vx->n;

	for(i=4; i<n; i+=5)
	{
		t1  = *(x++);
		p1  = t1*t1;
		s1 += p1;

		t2  = *(x++);
		p2  = t2*t2;
		s2 += p2;

		t3  = *(x++);
		p3  = t3*t3;
		s3 += p3;

		t4  = *(x++);
		p4  = t4*t4;
		s4 += p4;

		t5  = *(x++);
		p5  = t5*t5;
		s5 += p5;
	}
	dot = s1+s2+s3+s4+s5;
	for(i-=4;i<n;i++)
	{
		t    = *(x++);
		dot += t*t;
	}
	*val = sqrt(dot);
	return LIS_SUCCESS;
}

